import { getMessage } from '../utils/util.js';
import logger from '../utils/logger.js';
export function createPath(path, config) {
    const concatenatedPath = config.prependPath + path;
    try {
        return config.rewrite(concatenatedPath);
    }
    catch (err) {
        logger.warn(`Could not rewrite path "${concatenatedPath}". Falling back to original path`);
        return concatenatedPath;
    }
}
export function createHeaders(req, config) {
    const headers = {};
    // forwarded request falls back to http 1.x, so we must have a valid host header
    if (config.changeOrigin) {
        const { hostname, port } = config.url;
        headers.host = port ? `${hostname}:${port}` : hostname;
    }
    else {
        const authority = req.headers[':authority'];
        headers.host = authority || req.headers.host;
    }
    return headers;
}
export async function pipe(proxyRes, res) {
    return new Promise(resolve => {
        const { statusCode = 0 } = proxyRes;
        let bytesWritten = 0;
        function setHeaders() {
            if (!bytesWritten) {
                res.statusCode = statusCode;
                for (const [key, value] of Object.entries(proxyRes.headers)) {
                    res.setHeader(key, value);
                }
            }
        }
        function onDrain() {
            proxyRes.resume();
        }
        // pipe the response data from proxyRes to res
        proxyRes
            .on('data', (buf) => {
            setHeaders();
            bytesWritten += buf.length;
            // TS is not able to infer the `write` method from both response classes (http/http2)
            if (!res.write(buf)) {
                proxyRes.pause();
            }
        })
            .on('end', () => {
            try {
                setHeaders();
            }
            catch (err) {
                logger.error(`Could not proxy headers: ${getMessage(err)}`);
            }
            // trailers are only there, when version is compliant
            // proxyRes.trailers contains headers that are sent at the end of the request
            if (proxyRes.trailers)
                res.addTrailers(proxyRes.trailers);
            res.end();
            resolve();
        })
            .on('close', () => {
            res.off('drain', onDrain);
        });
        res.on('drain', onDrain);
    });
}
