import proxy from 'http2-proxy';
import createError from 'http-errors';
import zlib from 'zlib';
import { pipe } from './http.js';
function decompress(stream, decompressor) {
    stream.pipe(decompressor);
    return decompressor;
}
function extendRequestHeaders(reqOpts, options) {
    const headers = reqOpts.headers ??= {};
    Object.assign(headers, options.headers);
    for (const key of Object.keys(headers)) {
        if (headers[key] === undefined || headers[key] === null)
            delete headers[key];
    }
    return headers;
}
export function interceptData(req, res, options) {
    const { protocol, hostname, port } = options.url;
    let buffer = '';
    return new Promise((resolve, reject) => {
        // eslint-disable-next-line @typescript-eslint/no-floating-promises
        proxy.web(
        // force using http2 signature of `proxy.web` for additional options
        req, res, {
            protocol: protocol, // needed to satisfy type checker
            hostname,
            port: parseInt(port, 10), // needed to satisfy type checker
            rejectUnauthorized: options.secure,
            onReq(_req, reqOpts) {
                const headers = extendRequestHeaders(reqOpts, options);
                headers['accept-encoding'] = 'gzip, deflate';
                return Promise.resolve(); // needed to satisfy type checker
            },
            onRes(_req, _res, proxyRes_) {
                // typing bug in http2-proxy? `proxyRes` is an `IncomingMessage` at runtime
                const proxyRes = proxyRes_;
                if (proxyRes.statusCode >= 400) {
                    reject(createError(proxyRes.statusCode, proxyRes.statusMessage || ''));
                    return Promise.resolve(); // needed to satisfy type checker
                }
                const headers = proxyRes.headers;
                let transform;
                switch (headers['content-encoding']) {
                    case 'gzip':
                        transform = decompress(proxyRes, zlib.createGunzip());
                        break;
                    case 'deflate':
                        transform = decompress(proxyRes, zlib.createInflate());
                        break;
                    case 'br':
                        transform = decompress(proxyRes, zlib.createBrotliDecompress());
                        break;
                    default: transform = proxyRes;
                }
                transform
                    .on('error', reject)
                    .on('data', (buf) => {
                    buffer += buf.toString();
                })
                    .on('end', () => {
                    const contentType = headers['content-type'];
                    if (typeof contentType === 'string' && contentType.startsWith('application/json')) {
                        return resolve(JSON.parse(buffer));
                    }
                    resolve(buffer);
                });
                return Promise.resolve(); // needed to satisfy type checker
            }
        });
    });
}
export async function all(req, res, configs) {
    // TODO: property "readableEnded" is not available in type `ServerResponse` (but in `IncomingMessage`)
    if (req.aborted || res.readableEnded)
        return;
    return Promise.all(configs.map(config => interceptData(req, res, config)));
}
export async function web(req, res, options) {
    // TODO: property "readableEnded" is not available in type `ServerResponse` (but in `IncomingMessage`)
    if (req.aborted || res.readableEnded)
        return;
    const { protocol, hostname, port } = options.url;
    return proxy.web(
    // force using http2 signature of `proxy.web` for additional options
    req, res, {
        protocol: protocol, // needed to satisfy type checker
        hostname,
        port: parseInt(port, 10), // needed to satisfy type checker
        path: options.path,
        // We need to increase the timeout, because the middleware might take a while to respond
        proxyTimeout: 30000,
        rejectUnauthorized: options.secure,
        async onReq(req, reqOpts, cb) {
            extendRequestHeaders(reqOpts, options);
            await options.onReq?.(req, reqOpts, cb);
        },
        async onRes(req, res, proxyRes_, cb) {
            // typing bug in http2-proxy? `proxyRes` is an `IncomingMessage` at runtime
            const proxyRes = proxyRes_;
            await options.onRes?.(req, res, proxyRes, cb);
            await pipe(proxyRes, res);
        }
    });
}
export async function ws(req, socket, head, options) {
    const { protocol, hostname, port } = options.url;
    return proxy.ws(
    // force using http2 signature of `proxy.ws` for additional options
    req, socket, head, {
        protocol: protocol, // needed to satisfy type checker
        hostname,
        port: parseInt(port, 10), // needed to satisfy type checker
        path: options.path,
        rejectUnauthorized: options.secure,
        onReq(_req, reqOpts) {
            extendRequestHeaders(reqOpts, options);
            return Promise.resolve(); // needed to satisfy type checker
        }
    });
}
