import colors from 'picocolors';
import { PROJECT_NAME } from './utils/util.js';
import logger from './utils/logger.js';
import { createHttpProxy } from './http/http-proxy.js';
import detectDevServerMiddleware from './middlewares/detect-dev-server.js';
import manifestMiddleware from './middlewares/manifest.js';
import proxyMiddleware from './middlewares/proxy.js';
import proxyFrontendsMiddleware from './middlewares/proxy-frontends.js';
import upgradeMiddleware from './middlewares/upgrade.js';
const DEF_LOG_LEVEL = (process.env.NODE_ENV === 'test') ? 'error' : 'warn';
export default function pluginProxy(options) {
    let resolvedConfig;
    logger.level = options?.logLevel ?? DEF_LOG_LEVEL;
    // filter frontends without target
    const frontendOptions = (options?.frontends || []).filter(frontend => {
        const hasTarget = !!frontend.target;
        if (!hasTarget)
            logger.warn('You have to provide a target for every frontend!');
        return hasTarget;
    });
    return {
        name: PROJECT_NAME,
        enforce: 'pre',
        configResolved(config) {
            resolvedConfig = config;
        },
        configureServer(server) {
            if (options?.httpProxy) {
                const { target, port = 8080 } = options.httpProxy;
                // overwrite printUrls from vite
                const printUrls = server.printUrls.bind(server);
                server.printUrls = function () {
                    printUrls();
                    console.log(`  ${colors.green('➜')}  Http proxy: ${colors.cyan(`http://localhost:${port}/`)}`);
                };
                createHttpProxy(target, port);
            }
            // resolve proxy configurations
            const webProxies = [];
            const wsProxies = [];
            if (options?.proxy) {
                for (const [key, proxy] of Object.entries(options.proxy)) {
                    const { prependPath = '', changeOrigin = false, secure = true } = proxy;
                    const url = new URL(proxy.target);
                    const condition = key.startsWith('^') ? new RegExp(key) : key;
                    const rewrite = proxy.rewrite ?? (path => path);
                    const proxies = proxy.ws ? wsProxies : webProxies;
                    proxies.push({ url, secure, condition, prependPath, rewrite, changeOrigin });
                }
            }
            // resolve frontend configurations
            const frontends = frontendOptions.map(frontend => {
                const { path = '', secure = true } = frontend;
                const url = new URL(frontend.target);
                const condition = path.startsWith('^') ? new RegExp(path) : path;
                return { url, secure, condition };
            });
            // install all middlewares
            server.middlewares.use(detectDevServerMiddleware(server, frontends));
            server.middlewares.use(manifestMiddleware(frontends, resolvedConfig.base));
            server.middlewares.use(proxyMiddleware(webProxies));
            server.middlewares.use(proxyFrontendsMiddleware(frontends, resolvedConfig.base));
            server.httpServer?.on('upgrade', upgradeMiddleware(wsProxies));
        }
    };
}
