import { join } from 'node:path';
import fs from 'node:fs/promises';
import parseurl from 'parseurl';
import finalhandler from 'finalhandler';
import createError, { HttpError } from 'http-errors';
import { getStack, shouldHandle } from '../utils/util.js';
import logger from '../utils/logger.js';
import { pipe } from '../http/http.js';
import * as proxy from '../http/proxy.js';
function proxyRequest(req, res, frontend) {
    return proxy.web(req, res, {
        url: frontend.url,
        secure: frontend.secure,
        headers: { 'proxied-by-ox-proxy': 'true' },
        async onRes(_req, res, proxyRes_) {
            // typing bug in http2-proxy? `proxyRes` is an `IncomingMessage` at runtime
            const proxyRes = proxyRes_;
            if (proxyRes.statusCode >= 400) {
                throw createError(proxyRes.statusCode, proxyRes.statusMessage || '');
            }
            /* if (proxyRes.headers['content-length'] && proxyRes.headers['content-length'] === '0') {
              throw createError(proxyRes.statusCode, 'Wrong frontend configuration.')
            } */
            await pipe(proxyRes, res);
        }
    });
}
export default function proxyFrontendsMiddleware(frontends, base = '/') {
    const cache = new Map();
    // this is used to remove the /@fs/ prefix from the path
    const fsPathLength = process.platform === 'win32' ? '/@fs/'.length : '/@fs'.length;
    return async (req, res, next) => {
        if (req.headers['proxied-by-ox-proxy'])
            return next();
        const { path, pathname } = parseurl(req) ?? {};
        if (!path || !pathname)
            return next();
        // five stage system
        // 1. if we have a cached path, we can directly request the file from that system
        // 2. if it starts with /@fs, we can forward the request to the dev-server with according allowed paths
        // 3. if we have multiple dev-servers, we can check if the file exists there in the root
        // 4. check, if a path is configured for frontends and try to find the file there
        // 5. otherwise, request them in order
        // stage 1
        if (cache.has(path)) {
            logger.debug(`Found a handler for ${path} in cache`);
            try {
                const frontend = cache.get(path);
                await proxyRequest(req, res, frontend);
                logger.debug(`Delivered ${path} by a cached handler`);
                return;
            }
            catch (err) {
                logger.warn(`The handler for ${path} could not be used anymore.`);
                cache.delete(path);
            }
        }
        // stage 2
        if (path.startsWith(`${base}@fs/`)) {
            for (const frontend of frontends) {
                if (!frontend?.config?.fs?.allow)
                    continue;
                const fsPath = path.replace(base, '/').substring(fsPathLength);
                const allowed = frontend.config.fs.allow.some(allowed => fsPath.startsWith(allowed));
                if (!allowed)
                    continue;
                try {
                    logger.info(`Proxy ${path} to frontend ${frontend.url.origin}`);
                    await proxyRequest(req, res, frontend);
                    cache.set(path, frontend);
                    logger.debug(`Found ${path} at ${frontend.url.origin} in stage 2`);
                    return;
                }
                catch (err) {
                    logger.debug(`Did not find ${path} at ${frontend.url.origin}, try next.`);
                }
            }
        }
        // stage 3
        for (const frontend of frontends) {
            if (!frontend?.config?.root)
                continue;
            const file = join(frontend.config.root, pathname === base ? '/index.html' : pathname.replace(base, ''));
            try {
                await fs.stat(file);
                // rewrite the url to the frontend base
                req.url = req.url?.replace(base, frontend.config.base || '/') ?? req.url;
                req.originalUrl = req.originalUrl?.replace(base, frontend.config.base || '/') ?? req.originalUrl;
                logger.info(`Proxy ${path} to frontend ${frontend.url.origin} after fs.stat`);
                await proxyRequest(req, res, frontend);
                cache.set(path, frontend);
                logger.debug(`Found ${path} at ${frontend.url.origin} in stage 3`);
                return;
            }
            catch (err) {
                logger.info(`Could not find ${path} from ${frontend.url.origin} at ${frontend.config.root}`);
            }
        }
        // stage 4
        for (const frontend of frontends) {
            if (!frontend.condition || !shouldHandle(frontend.condition, path))
                continue;
            logger.info(`Proxy ${path} to frontend ${frontend.url.origin}`);
            try {
                await proxyRequest(req, res, frontend);
                cache.set(path, frontend);
                logger.debug(`Found ${path} at ${frontend.url.origin} in stage 4`);
                return;
            }
            catch (err) {
                logger.debug(`Did not find ${path} at ${frontend.url.origin}, try next.`);
            }
        }
        // stage 5
        logger.debug(`Fall back to autofind ${path} in frontends`);
        const errors = [];
        for (const frontend of frontends) {
            try {
                await proxyRequest(req, res, frontend);
                cache.set(path, frontend);
                logger.debug(`Found ${path} at ${frontend.url.origin} in stage 5`);
                return;
            }
            catch (err) {
                logger.debug(`Did not find ${path} at ${frontend.url.origin}, try next.`);
                errors.push(err);
            }
        }
        const err = errors.reduce((memo, err) => {
            // keep the first non-http error and forward it
            if (!(memo instanceof HttpError))
                return memo;
            // otherwise, use the last http error
            return err;
        }, new createError.NotFound());
        const handle = finalhandler(req, res);
        if (err instanceof HttpError) {
            // in case of http-error the proxy explicitly received an error code
            logger.warn(`Could not find ${path} in frontends`);
            return handle(err);
        }
        logger.error(`Something went wrong, while resolving a file in frontends: ${getStack(err)}`);
        return handle();
    };
}
